<html>
    <head>
        <script src="/static/js/web3.js"></script>
        <script src="/static/js/contracts.js"></script>
    </head>
    <body>
        This is the default <code>index.html</code> template provided by
        <strong>populus.</strong>
    </body>
</html>
